


function [x lambda f0 df0 c dc]=primal_dual( x,f0_funhandle,fn_funhandle,dualhess_funhandle,niters,tol,display_iter,varargin )
% function [x lambda f0 df0 d dc]=primal_dual( x,f0_funhandle,fn_funhandle,dualhess_funhandle,niters,tol,display_iter,compute_step_bounds_funhandle,lambda_ini )

flag_stepbounds=0;
if nargin>7 && ~isempty(varargin{1,1})
    flag_stepbounds=1;
    compute_step_bounds_funhandle=varargin{1,1};
end

% parameters affecting convergence (rule of thumb: For aggressive/robust convergence, chose maxsdgbetafac=10/2)
maxsdgbetafac=1.0;
minsdgbetafac=1.0;

[c d e f clim]=fn_funhandle(x);
ind_pos=find(c>0);
if ~isempty(ind_pos)
    % error('Initial guess is not feasible.');
    alpha=min( sqrt(clim./(c+clim)) );
    x=0.9*alpha*x;  % this guarantees that the starting is feasible, which should accelerate MLS convergence
    [c d e f clim]=fn_funhandle(x);
    if display_iter
        fprintf('Initial guess no feasible, scaling by %e ensures all constraints are negatives (max(c)=%e)\n',alpha,max(c));
    end
    if max(c)>0
        error('Initial guess infeasible.');
    end
end

% initial guess of the Lagrange multipliers
if nargin>8 && ~isempty(varargin{1,2})
    lambda=varargin{1,2};
else
    lambda=-1./c;  
end

sdg=-sum(c.*lambda,1);  % surrogate duality gap
sdgbetafac=maxsdgbetafac;  % sdg update factor 
sdgbeta=sdg/sdgbetafac/size(c,1);  % sdg update

for newtoniter=1:niters
    
    % newton search direction
    [dir_x dir_lambda]=compute_newton_search_direction(x,lambda,f0_funhandle,fn_funhandle,dualhess_funhandle,sdgbeta);
    
    % step size
    if flag_stepbounds
        step=ls_backtracking(x,lambda,dir_x,dir_lambda,sdgbeta,0.01,0.5,fn_funhandle,f0_funhandle,compute_step_bounds_funhandle);
    else
        step=ls_backtracking(x,lambda,dir_x,dir_lambda,sdgbeta,0.01,0.5,fn_funhandle,f0_funhandle);
    end    
    
    % update solution and Lagrange multipliers
    x=x + step*dir_x;
    lambda=lambda + step*dir_lambda;
    
    % compute residual
    [c dc]=fn_funhandle(x);
    [f0 df0]=f0_funhandle(x);
    [res_x res_lambda]=compute_residual(f0,df0,c,dc,lambda,sdgbeta);
    res_norm=norm(res_x);
    
    % update surrogate duality gap
    sdgbetafac=minsdgbetafac*exp(step*log(maxsdgbetafac/minsdgbetafac));
    sdg=-sum(c.*lambda,1);
    sdgbeta=sdg/sdgbetafac/size(c,1);
    
    [cmax imax]=max(c);
    if display_iter
        fprintf('\tNewton it. %d out of %d: f0(x)=%3.3e  max_const=%3.3e  max_const_ind=%d  max_const_lambda=%3.3e  step=%3.3e  sdg=%3.3e  beta_sdg=%3.3e  beta_sdg_fac=%3.3e  res_norm=%3.3e\n', ...
            newtoniter,niters,f0,cmax,imax,lambda(imax),step,sdg,sdgbeta,sdgbetafac,res_norm);
    end
    
    % convergence    
    if ( res_norm<tol && (sdg<tol || step<1e-8) )
      break;
    end
    
end



function [dir_x dir_lambda]=compute_newton_search_direction(x,lambda,f0_funhandle,fn_funhandle,dualhess_funhandle,sdgbeta)
% function [dir_x dir_lambda]=compute_newton_search_direction(x,lambda,f0_funhandle,fn_funhandle,dualhess_funhandle,sdgbeta)

[c dc]=fn_funhandle(x);
[f0 df0]=f0_funhandle(x);

% blocks of the Hessian
hess_ul=dualhess_funhandle(x,lambda);  % upper left block
hess_ur=dc;  % upper right block
hess_ll=-diag(lambda)*(dc.');  % lower left block
hess_lr=-c;  % lower right block

hess_ur=hess_ur*diag(1./hess_lr);

% Schur complement
schur=hess_ul - hess_ur*hess_ll;

% residual
[res_x res_lambda]=compute_residual(f0,df0,c,dc,lambda,sdgbeta);

% search direction for the variables
tmp1=hess_ur*res_lambda - res_x;

% dir_x=schur\tmp1;  % solve schur*x=tmp1
dir_x=pinv(schur)*tmp1;  % solve schur*x=tmp1 -- slower but more stable

% search direction for the Lagrange multipliers
dir_lambda=( -hess_ll*dir_x-res_lambda )./hess_lr;



function [res_x res_lambda]=compute_residual(f,df,c,dc,lambda,sdgbeta)
% function [res_x res_lambda]=compute_residual(f,df,c,dc,lambda,sdgbeta)

res_x=df + dc*lambda;
res_lambda=-lambda.*c - sdgbeta;



function t=ls_backtracking(x0,lambda0,dir_x,dir_lambda,sdgbeta,alpha,beta,fn_funhandle,f0_funhandle,varargin)
% function ls_backtracking(x0,lambda0,dir_x,dir_lambda,sdgbeta,alpha,beta,fn_funhandle,f0_funhandle,compute_step_bounds_funhandle)

% maximum step size still yielding positive Lagrange multipliers
tmax=1e5;
ind_neg=find(dir_lambda<0);
if ~isempty(ind_neg)
    tmax=min( -lambda0(ind_neg)./dir_lambda(ind_neg) );
end
tmax=min(1,tmax);

if nargin>9
    % user-defined computation of the steps bounds
    compute_step_bounds_funhandle=varargin{1,1};
    tmax=compute_step_bounds_funhandle(x0,dir_x,tmax);
else
    % decrease tmax until feasible point (i.e., all constraints are negative)
    for j=1:1000
        x=x0 + tmax*dir_x;
        c=fn_funhandle(x);
        ind_pos=find(c>0);
        if isempty(ind_pos)
            break;
        else
            tmax=tmax*beta;
        end
    end
    
    if j==1000
        error('Impossible to find a step size ensuring feasibility.');
    end
end

t=tmax*0.99;

% backtracking line-search
[c dc]=fn_funhandle(x0);
[f0 df0]=f0_funhandle(x0);
[res_x res_lambda]=compute_residual(f0,df0,c,dc,lambda0,sdgbeta);
res_norm_ini=norm([ res_x;res_lambda ]);
for j=1:50
    
    % update solution and Lagrange multiplirs
    x=x0 + t*dir_x;
    lambda=lambda0 + t*dir_lambda;
    
    % recompute the residual at that new solution
    [c dc]=fn_funhandle(x);
    [f0 df0]=f0_funhandle(x);
    [res_x res_lambda]=compute_residual(f0,df0,c,dc,lambda,sdgbeta);
    res_norm=norm([ res_x;res_lambda ]);
    
    % convergence criterion/step size update
    if res_norm<=( 1.0-alpha*t )*res_norm_ini
        break;
    else
        t=t*beta;
    end
  
end

if j==50
    % error('Backtracking did not converge.');
    t=-0.1;
end





    






























